import sys
from optparse import NO_DEFAULT, OptionParser
from django.core.management.base import CommandError, BaseCommand, handle_default_options
from django.utils.importlib import import_module


class Command(BaseCommand):
    help = "A wrapper for Avocado subcommands"

    commands = {
        'init': 'init',
        'check': 'check',
        'data': 'data',
        'cache': 'cache',
        'legacy': 'legacy',
        'lexicon': 'lexicon',
        'history': 'history',
        'migration': 'migration',
    }

    def print_subcommands(self, prog_name):
        usage = ['', 'Available subcommands:']
        for name in sorted(self.commands.keys()):
            usage.append('  {0}'.format(name))
        return '\n'.join(usage)

    def usage(self, subcommand):
        usage = '%prog {0} subcommand [options] [args]'.format(subcommand)
        if self.help:
            return '{0}\n\n{1}'.format(usage, self.help)
        return usage

    def print_help(self, prog_name, subcommand):
        super(Command, self).print_help(prog_name, subcommand)
        sys.stdout.write('{0}\n\n'.format(self.print_subcommands(prog_name)))

    def get_subcommand(self, name):
        try:
            module = import_module('avocado.management.subcommands.{0}'.format(name))
            return module.Command()
        except KeyError:
            raise CommandError('Unknown subcommand: avocado {0}'.format(name))

    def run_from_argv(self, argv):
        """Set up any environment changes requested (e.g., Python path
        and Django settings), then run this command.
        """
        if len(argv) > 2 and not argv[2].startswith('-') and argv[2] in self.commands.keys():
            subcommand = self.commands[argv[2]]
            klass = self.get_subcommand(subcommand)
            parser = OptionParser(prog=argv[0], usage=klass.usage('{0} {1}'.format(argv[1], subcommand)),
                version=klass.get_version(), option_list=klass.option_list)
            options, args = parser.parse_args(argv[3:])
            args = [subcommand] + args
        else:
            parser = self.create_parser(argv[0], argv[1])
            options, args = parser.parse_args(argv[2:])

        handle_default_options(options)
        self.execute(*args, **options.__dict__)

    def handle(self, *args, **options):
        if not args or args[0] not in self.commands.keys():
            return self.print_help('./manage.py', 'avocado')
        subcommand, args = args[0], args[1:]

        klass = self.get_subcommand(subcommand)
        # Grab out a list of defaults from the options. optparse does this for
        # us when the script runs from the command line, but since
        # call_command can be called programatically, we need to simulate the
        # loading and handling of defaults (see #10080 for details).
        defaults = {}
        for opt in klass.option_list:
            if opt.default is NO_DEFAULT:
                defaults[opt.dest] = None
            else:
                defaults[opt.dest] = opt.default
        defaults.update(options)

        return klass.execute(*args, **defaults)



from django.db import models
from django.core.cache import cache
from django.db.models.query import QuerySet

NEVER_EXPIRE = 60 * 60 * 24 * 30 # 30 days
INSTANCE_CACHE_KEY = '{0}.{1}:{2}'


def instance_cache_key(instance, label=None, version=None):
    "Creates a cache key for the instance with an optional label and token."
    opts = instance._meta
    key = INSTANCE_CACHE_KEY.format(opts.app_label, opts.module_name, instance.pk)
    if label:
        key = '{0}-{1}'.format(key, label)
    if version:
        key = '{0}-{1}'.format(key, version)
    return key


def cached_property(label, version=None, timeout=NEVER_EXPIRE):
    "Wraps a function and caches the output indefinitely."
    def decorator(func):
        def wrapped(self):
            # Do not cache non-persisted objects
            if not self.pk:
                return func(self)

            # If this is a function, pass `label' and `self` in as arguments
            if callable(version):
                _version = version(self, label=label)
            else:
                _version = getattr(self, version)
                # Call if a method
                if callable(_version):
                    _version = _version()
            # If no version is defined, the cache cannot be reliably stored
            if _version is None:
                data = func(self)
            else:
                key = instance_cache_key(self, label=label, version=_version)
                data = cache.get(key)
                if data is None:
                    data = func(self)
                    # Don't bother caching if the data is None
                    if data is not None:
                        cache.set(key, data, timeout=timeout)
            return data
        return property(wrapped)
    return decorator


def post_save_cache(sender, instance, **kwargs):
    """General post-save handler for caching model instances. NOTE: This must
    be used in conjunction with the `pre_delete_uncache` since the cache is set
    to never expire.
    """
    cache.set(instance_cache_key(instance), instance, timeout=NEVER_EXPIRE)


def pre_delete_uncache(sender, instance, **kwargs):
    "General post-delete handler for removing cache for model instances."
    cache.delete(instance_cache_key(instance))


# `pk` is used as an alias, so this is constant
PK_LOOKUPS = ['pk', 'pk__exact']


class CacheQuerySet(QuerySet):
    def filter(self, *args, **kwargs):
        """For primary-key-based lookups, instances may be cached to prevent
        excessive database hits. If this is a primary-key lookup, the cache
        will be checked and populate the `_result_cache` if available.
        """
        clone = super(CacheQuerySet, self).filter(*args, **kwargs)

        pk = None
        opts = self.model._meta
        pk_name = opts.pk.name

        # Look for `pk` and the actual name of the primary key field
        for key in PK_LOOKUPS + [pk_name, '{0}__exact'.format(pk_name)]:
            if key in kwargs:
                pk = kwargs[key]
                break

        if pk is not None:
            key = INSTANCE_CACHE_KEY.format(opts.app_label, opts.module_name, pk)
            obj = cache.get(key)
            if obj is not None:
                clone._result_cache = [obj]

        return clone


class CacheManager(models.Manager):
    def get_query_set(self):
        return CacheQuerySet(self.model)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'DataCategory'
        db.create_table('avocado_datacategory', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=50, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('keywords', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('archived', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('published', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('modified', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
            ('parent', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='children', null=True, to=orm['avocado.DataCategory'])),
            ('order', self.gf('django.db.models.fields.FloatField')(null=True, db_column='_order', blank=True)),
        ))
        db.send_create_signal('avocado', ['DataCategory'])

        # Adding model 'DataField'
        db.create_table('avocado_datafield', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=50, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('keywords', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('archived', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('published', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('modified', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
            ('name_plural', self.gf('django.db.models.fields.CharField')(max_length=60, null=True, blank=True)),
            ('app_name', self.gf('django.db.models.fields.CharField')(max_length=50)),
            ('model_name', self.gf('django.db.models.fields.CharField')(max_length=50)),
            ('field_name', self.gf('django.db.models.fields.CharField')(max_length=50)),
            ('unit', self.gf('django.db.models.fields.CharField')(max_length=30, null=True, blank=True)),
            ('unit_plural', self.gf('django.db.models.fields.CharField')(max_length=40, null=True, blank=True)),
            ('category', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['avocado.DataCategory'], null=True, blank=True)),
            ('enumerable', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('searchable', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('translator', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('data_modified', self.gf('django.db.models.fields.DateTimeField')(null=True)),
            ('data_source', self.gf('django.db.models.fields.CharField')(max_length=250, null=True, blank=True)),
            ('group', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='fields+', null=True, to=orm['auth.Group'])),
            ('order', self.gf('django.db.models.fields.FloatField')(null=True, db_column='_order', blank=True)),
        ))
        db.send_create_signal('avocado', ['DataField'])

        # Adding unique constraint on 'DataField', fields ['app_name', 'model_name', 'field_name']
        db.create_unique('avocado_datafield', ['app_name', 'model_name', 'field_name'])

        # Adding M2M table for field sites on 'DataField'
        db.create_table('avocado_datafield_sites', (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('datafield', models.ForeignKey(orm['avocado.datafield'], null=False)),
            ('site', models.ForeignKey(orm['sites.site'], null=False))
        ))
        db.create_unique('avocado_datafield_sites', ['datafield_id', 'site_id'])

        # Adding model 'DataConcept'
        db.create_table('avocado_dataconcept', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=50, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('keywords', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('archived', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('published', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('modified', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
            ('name_plural', self.gf('django.db.models.fields.CharField')(max_length=60, null=True, blank=True)),
            ('category', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['avocado.DataCategory'], null=True, blank=True)),
            ('group', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='concepts+', null=True, to=orm['auth.Group'])),
            ('order', self.gf('django.db.models.fields.FloatField')(null=True, db_column='_order', blank=True)),
            ('formatter', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('queryview', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
        ))
        db.send_create_signal('avocado', ['DataConcept'])

        # Adding M2M table for field sites on 'DataConcept'
        db.create_table('avocado_dataconcept_sites', (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('dataconcept', models.ForeignKey(orm['avocado.dataconcept'], null=False)),
            ('site', models.ForeignKey(orm['sites.site'], null=False))
        ))
        db.create_unique('avocado_dataconcept_sites', ['dataconcept_id', 'site_id'])

        # Adding model 'DataConceptField'
        db.create_table('avocado_dataconceptfield', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('field', self.gf('django.db.models.fields.related.ForeignKey')(related_name='concept_fields', to=orm['avocado.DataField'])),
            ('concept', self.gf('django.db.models.fields.related.ForeignKey')(related_name='concept_fields', to=orm['avocado.DataConcept'])),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('name_plural', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('order', self.gf('django.db.models.fields.FloatField')(null=True, db_column='_order', blank=True)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('modified', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
        ))
        db.send_create_signal('avocado', ['DataConceptField'])

        # Adding model 'DataContext'
        db.create_table('avocado_datacontext', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=50, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('keywords', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('archived', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('published', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('modified', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
            ('json', self.gf('jsonfield.fields.JSONField')(default={}, null=True, blank=True)),
            ('session', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('composite', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('count', self.gf('django.db.models.fields.IntegerField')(null=True, db_column='_count')),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='datacontext+', null=True, to=orm['auth.User'])),
            ('session_key', self.gf('django.db.models.fields.CharField')(max_length=40, null=True, blank=True)),
        ))
        db.send_create_signal('avocado', ['DataContext'])

        # Adding model 'DataView'
        db.create_table('avocado_dataview', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=50, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('keywords', self.gf('django.db.models.fields.CharField')(max_length=100, null=True, blank=True)),
            ('archived', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('published', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('modified', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, blank=True)),
            ('json', self.gf('jsonfield.fields.JSONField')(default={}, null=True, blank=True)),
            ('session', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('count', self.gf('django.db.models.fields.IntegerField')(null=True, db_column='_count')),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='dataview+', null=True, to=orm['auth.User'])),
            ('session_key', self.gf('django.db.models.fields.CharField')(max_length=40, null=True, blank=True)),
        ))
        db.send_create_signal('avocado', ['DataView'])


    def backwards(self, orm):
        # Removing unique constraint on 'DataField', fields ['app_name', 'model_name', 'field_name']
        db.delete_unique('avocado_datafield', ['app_name', 'model_name', 'field_name'])

        # Deleting model 'DataCategory'
        db.delete_table('avocado_datacategory')

        # Deleting model 'DataField'
        db.delete_table('avocado_datafield')

        # Removing M2M table for field sites on 'DataField'
        db.delete_table('avocado_datafield_sites')

        # Deleting model 'DataConcept'
        db.delete_table('avocado_dataconcept')

        # Removing M2M table for field sites on 'DataConcept'
        db.delete_table('avocado_dataconcept_sites')

        # Deleting model 'DataConceptField'
        db.delete_table('avocado_dataconceptfield')

        # Deleting model 'DataContext'
        db.delete_table('avocado_datacontext')

        # Deleting model 'DataView'
        db.delete_table('avocado_dataview')


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        'avocado.datacategory': {
            'Meta': {'ordering': "('order', 'name')", 'object_name': 'DataCategory'},
            'archived': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50', 'null': 'True', 'blank': 'True'}),
            'order': ('django.db.models.fields.FloatField', [], {'null': 'True', 'db_column': "'_order'", 'blank': 'True'}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'children'", 'null': 'True', 'to': "orm['avocado.DataCategory']"}),
            'published': ('django.db.models.fields.BooleanField', [], {'default': 'False'})
        },
        'avocado.dataconcept': {
            'Meta': {'ordering': "('order',)", 'object_name': 'DataConcept'},
            'archived': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'category': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['avocado.DataCategory']", 'null': 'True', 'blank': 'True'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'fields': ('django.db.models.fields.related.ManyToManyField', [], {'related_name': "'concepts'", 'symmetrical': 'False', 'through': "orm['avocado.DataConceptField']", 'to': "orm['avocado.DataField']"}),
            'formatter': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'group': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'concepts+'", 'null': 'True', 'to': "orm['auth.Group']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50', 'null': 'True', 'blank': 'True'}),
            'name_plural': ('django.db.models.fields.CharField', [], {'max_length': '60', 'null': 'True', 'blank': 'True'}),
            'order': ('django.db.models.fields.FloatField', [], {'null': 'True', 'db_column': "'_order'", 'blank': 'True'}),
            'published': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'queryview': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'sites': ('django.db.models.fields.related.ManyToManyField', [], {'symmetrical': 'False', 'related_name': "'concepts+'", 'blank': 'True', 'to': "orm['sites.Site']"})
        },
        'avocado.dataconceptfield': {
            'Meta': {'ordering': "('order',)", 'object_name': 'DataConceptField'},
            'concept': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'concept_fields'", 'to': "orm['avocado.DataConcept']"}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'field': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'concept_fields'", 'to': "orm['avocado.DataField']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'name_plural': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'order': ('django.db.models.fields.FloatField', [], {'null': 'True', 'db_column': "'_order'", 'blank': 'True'})
        },
        'avocado.datacontext': {
            'Meta': {'object_name': 'DataContext'},
            'archived': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'composite': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'count': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'db_column': "'_count'"}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'json': ('jsonfield.fields.JSONField', [], {'default': '{}', 'null': 'True', 'blank': 'True'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50', 'null': 'True', 'blank': 'True'}),
            'published': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'session': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'session_key': ('django.db.models.fields.CharField', [], {'max_length': '40', 'null': 'True', 'blank': 'True'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'datacontext+'", 'null': 'True', 'to': "orm['auth.User']"})
        },
        'avocado.datafield': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_name', 'model_name', 'field_name'),)", 'object_name': 'DataField'},
            'app_name': ('django.db.models.fields.CharField', [], {'max_length': '50'}),
            'archived': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'category': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['avocado.DataCategory']", 'null': 'True', 'blank': 'True'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'data_modified': ('django.db.models.fields.DateTimeField', [], {'null': 'True'}),
            'data_source': ('django.db.models.fields.CharField', [], {'max_length': '250', 'null': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'enumerable': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'field_name': ('django.db.models.fields.CharField', [], {'max_length': '50'}),
            'group': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'fields+'", 'null': 'True', 'to': "orm['auth.Group']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'model_name': ('django.db.models.fields.CharField', [], {'max_length': '50'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50', 'null': 'True', 'blank': 'True'}),
            'name_plural': ('django.db.models.fields.CharField', [], {'max_length': '60', 'null': 'True', 'blank': 'True'}),
            'order': ('django.db.models.fields.FloatField', [], {'null': 'True', 'db_column': "'_order'", 'blank': 'True'}),
            'published': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'searchable': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'sites': ('django.db.models.fields.related.ManyToManyField', [], {'symmetrical': 'False', 'related_name': "'fields+'", 'blank': 'True', 'to': "orm['sites.Site']"}),
            'translator': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'unit': ('django.db.models.fields.CharField', [], {'max_length': '30', 'null': 'True', 'blank': 'True'}),
            'unit_plural': ('django.db.models.fields.CharField', [], {'max_length': '40', 'null': 'True', 'blank': 'True'})
        },
        'avocado.dataview': {
            'Meta': {'object_name': 'DataView'},
            'archived': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'count': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'db_column': "'_count'"}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'json': ('jsonfield.fields.JSONField', [], {'default': '{}', 'null': 'True', 'blank': 'True'}),
            'keywords': ('django.db.models.fields.CharField', [], {'max_length': '100', 'null': 'True', 'blank': 'True'}),
            'modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50', 'null': 'True', 'blank': 'True'}),
            'published': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'session': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'session_key': ('django.db.models.fields.CharField', [], {'max_length': '40', 'null': 'True', 'blank': 'True'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'dataview+'", 'null': 'True', 'to': "orm['auth.User']"})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'sites.site': {
            'Meta': {'ordering': "('domain',)", 'object_name': 'Site', 'db_table': "'django_site'"},
            'domain': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        }
    }

    complete_apps = ['avocado']


import csv
from _base import BaseExporter


class CSVExporter(BaseExporter):
    short_name = 'CSV'
    long_name = 'Comma-Separated Values (CSV)'

    file_extension = 'csv'
    content_type = 'text/csv'

    preferred_formats = ('csv', 'number', 'string')

    def write(self, iterable, buff=None, *args, **kwargs):
        header = []
        buff = self.get_file_obj(buff)
        writer = csv.writer(buff, quoting=csv.QUOTE_MINIMAL)

        for i, row_gen in enumerate(self.read(iterable, *args, **kwargs)):
            row = []
            for data in row_gen:
                if i == 0:
                    header.extend(data.keys())
                row.extend(data.values())
            if i == 0:
                writer.writerow(header)
            writer.writerow(row)
        return buff



from cStringIO import StringIO


class BaseExporter(object):
    "Base class for all exporters"
    file_extension = 'txt'
    content_type = 'text/plain'
    preferred_formats = []

    def __init__(self, concepts):
        self.concepts = concepts
        self.params = []
        self.row_length = 0

        for concept in concepts:
            length = concept.concept_fields.count()
            self.row_length += length
            self.params.append((concept.format, length))

    def get_file_obj(self, name=None):
        if name is None:
            return StringIO()
        if isinstance(name, basestring):
            return open(name, 'w+')
        return name

    def _format_row(self, row):
        for formatter, length in self.params:
            values, row = row[:length], row[length:]
            yield formatter(values, self.preferred_formats)

    def read(self, iterable, force_distinct=True, *args, **kwargs):
        """Takes an iterable that produces rows to be formatted.

        If `force_distinct` is true, rows will be filtered based on the slice
        of the row that is *up* for to be formatted. Note, this assumes the
        rows are ordered.
        """
        last_row = None
        for row in iterable:
            _row = row[:self.row_length]
            if force_distinct and _row == last_row:
                continue
            last_row = _row
            yield self._format_row(_row)

    def write(self, iterable, *args, **kwargs):
        raise NotImplemented



from django.db import models
from .cache import CacheQuerySet


# Taken from Carl Meyer's django-model-utils
class PassThroughManager(models.Manager):
    """Inherit from this Manager to enable you to call any methods from your
    custom QuerySet class from your manager. Simply define your QuerySet
    class, and return an instance of it from your manager's `get_query_set`
    method.

    Alternately, if you don't need any extra methods on your manager that
    aren't on your QuerySet, then just pass your QuerySet class to this
    class' constructer.

    class PostQuerySet(QuerySet):
        def enabled(self):
            return self.filter(disabled=False)

    class Post(models.Model):
        objects = PassThroughManager(PostQuerySet)

    """
    # pickling causes recursion errors
    _deny_methods = ['__getstate__', '__setstate__', '_db']

    def __init__(self, queryset_cls=None):
        self._queryset_cls = queryset_cls
        super(PassThroughManager, self).__init__()

    def __getattr__(self, name):
        if name in self._deny_methods:
            raise AttributeError(name)
        return getattr(self.get_query_set(), name)

    def get_query_set(self):
        if self._queryset_cls is not None:
            kwargs = {'model': self.model}
            if hasattr(self, '_db'):
                kwargs['using'] = self._db
            return self._queryset_cls(**kwargs)
        return super(PassThroughManager, self).get_query_set()


class PublishedQuerySet(CacheQuerySet):
    "Adds additional helper methods focused around access and permissions."
    def published(self):
        "Returns all published non-archived objects."
        return self.filter(published=True, archived=False)


class PublishedManager(PassThroughManager):
    def get_query_set(self):
        return PublishedQuerySet(self.model, using=self._db)



from django import forms
from django.db import models
from django.db.models.query import QuerySet
from django.core.exceptions import ValidationError
from modeltree.tree import trees
from avocado.core import loader
from avocado.conf import settings
from avocado.core.utils import get_form_class
from .operators import registry as operators


OPERATOR_MAP = settings.OPERATOR_MAP
INTERNAL_DATATYPE_FORMFIELDS = settings.INTERNAL_DATATYPE_FORMFIELDS


class Translator(object):
    """Given a `DataField` instance, a raw value and operator, a
    translator validates, cleans and constructs Django compatible
    lookups contained in the `query_modifiers` key in the output.

    The query modifiers are used to modify a `QuerySet` to filter
    the set down based on the raw input condition.
    """

    # An override of the list of supported operators for this particular
    # translator. This may be necessary for client programs which have
    # custom representations for certain fields.
    operators = None

    # Override of the field's default formfield class to be
    # used for validation. This is usually never necessary to override
    form_class = None

    def _parse_value(self, obj, key):
        """Handles parsing a value. This can be either a dict with a
        `value` and `label` key, some non-string iterable or the value
        itself.
        """
        if isinstance(obj, dict):
            return obj[key]
        if hasattr(obj, '__iter__'):
            return map(lambda x: self._parse_value(x, key), obj)
        return obj

    def _get_value(self, obj):
        return self._parse_value(obj, 'value')

    def _get_label(self, obj):
        return self._parse_value(obj, 'label')

    def get_operators(self, field):
        # Determine list of allowed operators
        return self.operators or OPERATOR_MAP[field.simple_type]

    def _validate_operator(self, field, uid, **kwargs):
        # Determine list of allowed operators
        allowed_operators = self.get_operators(field)

        # Special case for fields that are nullable
        if field.field.null:
            allowed_operators += ('isnull', '-isnull')

        # If uid is None, the default operator will be used
        uid = uid or allowed_operators[0]

        # Attempt to retrieve the operator.
        operator = operators.get(uid)

        # No operator is registered
        if operator is None:
            raise ValidationError('"{0}" is not a valid operator'.format(uid))

        # Ensure the operator is allowed
        if operator.uid not in allowed_operators:
            raise ValidationError('Operator "{0}" cannot be used for ' \
                'this translator'.format(operator))

        return operator

    def _validate_value(self, field, value, **kwargs):
        # If a form class is not specified, check to see if there is a custom
        # form_class specified for this datatype or if this translator has
        # one defined
        if 'form_class' not in kwargs:
            if self.form_class:
                kwargs['form_class'] = self.form_class
            elif field.simple_type in INTERNAL_DATATYPE_FORMFIELDS:
                name = INTERNAL_DATATYPE_FORMFIELDS[field.simple_type]
                kwargs['form_class'] = get_form_class(name)

        # The formfield is being used to clean the value, thus no
        # 'required' validation errors should be raised.
        kwargs['required'] = False

        # Special handling for primary keys
        if isinstance(field.field, models.AutoField):
            kwargs.pop('form_class', None)
            queryset = field.objects
            if hasattr(value, '__iter__'):
                formfield = forms.ModelMultipleChoiceField(queryset, **kwargs)
            else:
                formfield = forms.ModelChoiceField(queryset, **kwargs)
            cleaned_value = formfield.clean(value)
            return cleaned_value

        # If this field is flagged as enumerable, use a select multiple
        # by default.
        if field.enumerable and 'widget' not in kwargs:
            kwargs['widget'] = forms.SelectMultiple(choices=field.choices)

        # The model field instance has a convenience method called `formfield`
        # that is suited for the field type
        formfield = field.field.formfield(**kwargs)

        # Special case for ``None`` values since all form fields seem to handle
        # the conversion differently. Simply ignore the cleaning if ``None``,
        # this scenario occurs when a list of values are being queried and one
        # of them is to lookup NULL values. Note, the None is handled
        # downstream and is contained with the query directly.
        if hasattr(value, '__iter__'):
            cleaned_value = []
            for x in value:
                if x is not None:
                    cleaned_value.append(formfield.clean(x))
                # Django assumes an empty string when given a ``NoneType``
                # for char-based form fields, this is to ensure ``NoneType``
                # are passed through unmodified
                else:
                    cleaned_value.append(None)
            return cleaned_value
        return formfield.clean(value)

    def _get_not_null_pk(self, field, tree):
        """The below logic is required to get the expected results back
        when querying for NULL values. Performing a LEFT OUTER JOIN will
        cause non-existent rows on the right-hand side to be 'filled' in
        with NULL values. If the condition is explicitly querying for NULL
        values, this condition ensures this row actually exists.

        Django assumes in all cases when querying on a NULL value all joins
        in the chain up to that point must be promoted to LEFT OUTER JOINs,
        which could be a reasonable assumption for some cases, but for
        getting the existent rows of data back for our purposes, the
        assumption is wrong.
        """

        return tree.query_condition(field.model._meta.pk, 'isnull', False)

    def _condition(self, field, operator, value, tree):
        """Builds a `Q` object for `field` relative to `tree`.
        This handles a few edge cases such as passing a `None` in the list
        of values for an 'in' lookup and ensuring lookups for NULL values do
        not include filled in rows. Read more in `_get_not_null_pk`.
        """

        # Ensure this is a ModelTree instance
        tree = trees[tree]

        # Define condition
        condition = None

        # Flag for adding a condition with includes querying for a NULL value
        add_null = False

        value = self._normalize_value(field, value)

        # Assuming the operator and value validate, check for a NoneType value
        # if the operator is 'in'. This condition will be broken out into a
        # separate Q object
        if operator.lookup == 'exact' and value is None:
            add_null = True
        elif operator.lookup == 'isnull':
            add_null = True
        else:
            # Remove the None value from the list to process separately
            if operator.lookup == 'in':
                if None in value:
                    add_null = True
                    value.remove(None)

            # Process a normal value
            if value is not None:
                condition = tree.query_condition(field.field, operator.lookup, value)

            # Reset value to None for `null` processing
            value = None

        # The logical OR here is harmless since the only time where a previous
        # condition is defined is for an 'in' lookup.
        if add_null:
            # If value is None, this defaults to isnull=True
            if value is None:
                value = True
            # Read the _get_not_null_pk docs for more info
            null_condition = tree.query_condition(field.field, 'isnull', value)

            if field.model is not tree.root_model:
                null_condition = null_condition & \
                    self._get_not_null_pk(field, tree)

            # Tack on to the existing condition if one is defined
            if condition is not None:
                condition = condition | null_condition
            else:
                condition = null_condition

        if operator.negated:
            return ~condition
        return condition

    def _normalize_value(self, field, value):
        """Normalizes a cleaned value from some non-primitive type
        such as a model or queryset instance.
        """
        if field.simple_type == 'key':
            if isinstance(value, (list, tuple, QuerySet)):
                return [x.pk for x in value]
            return value.pk
        if isinstance(value, QuerySet):
            return [x.pk for x in value]
        if isinstance(value, models.Model):
            return value.pk
        return value

    def validate(self, field, operator, value, tree, **kwargs):
        value = self._get_value(value)
        operator = self._validate_operator(field, operator, **kwargs)

        # This is unique case since the operator is driving the required
        # type rather than using the datatype of the field. There is likely
        # a more elegant way to do this.
        if operator.lookup != 'isnull':
            value = self._validate_value(field, value, **kwargs)

        _value = self._normalize_value(field, value)
        if not operator.is_valid(_value):
            raise ValidationError('"{0}" is not valid for the operator '
                '"{1}"'.format(value, operator))

        return operator, value

    def language(self, field, operator, value, **kwargs):
        return u'{0} {1}'.format(field.name, operator.text(value))

    def translate(self, field, roperator, rvalue, tree, **kwargs):
        """Returns two types of queryset modifiers including:
            - the raw operator and value supplied
            - the validated and cleaned data
            - a Q object containing the condition (or multiple)
            - a dict of annotations to be used downstream

        It should be noted that no checks are performed to prevent the same
        name being used for annotations.
        """
        operator, value = self.validate(field, roperator, rvalue, tree, **kwargs)
        condition = self._condition(field, operator, value, tree)
        language = self.language(field, operator, value, **kwargs)

        return {
            'id': field.pk,
            'operator': roperator,
            'value': rvalue,
            'cleaned_data': {
                'value': value,
                'operator': operator,
                'language': language,
            },
            'query_modifiers': {
                'condition': condition,
                'annotations': None,
                'extra': None,
            }
        }


registry = loader.Registry(default=Translator)

# this will be invoked when it is imported by models.py to use the
# registry choices
loader.autodiscover('translators')



import re
from django.db import models
from django.contrib.sites.models import Site
from django.contrib.auth.models import User, Group
from django.utils.encoding import smart_unicode
from django.utils.translation import ugettext_lazy as _
from django.db.models.fields import FieldDoesNotExist
from django.db.models.signals import post_save, pre_delete
from django.core.validators import RegexValidator
from avocado.core import utils
from avocado.core.models import Base, BasePlural
from avocado.core.cache import (post_save_cache, pre_delete_uncache,
    cached_property)
from avocado.conf import settings
from avocado.managers import (DataFieldManager, DataConceptManager,
    DataCategoryManager)
from avocado.query.models import AbstractDataView, AbstractDataContext
from avocado.query.translators import registry as translators
from avocado.query.operators import registry as operators
from avocado.lexicon.models import Lexicon
from avocado.sets.models import ObjectSet
from avocado.stats.agg import Aggregator
from avocado.formatters import registry as formatters
from avocado.queryview import registry as queryviews


__all__ = ('DataCategory', 'DataConcept', 'DataField',
    'DataContext', 'DataView')


SIMPLE_TYPE_MAP = settings.SIMPLE_TYPE_MAP


ident_re = re.compile(r'^[a-zA-Z][a-zA-Z0-9_]*$')
validate_ident = RegexValidator(ident_re, _("Enter a valid 'identifier' " \
    "that is a valid Python variable name."), 'invalid')


class DataCategory(Base):
    "A high-level organization for data concepts."
    # A reference to a parent for hierarchical categories
    parent = models.ForeignKey('self', null=True, blank=True,
        related_name='children', limit_choices_to={'parent__isnull': True},
        help_text='Sub-categories are limited to one-level deep')
    order = models.FloatField(null=True, blank=True, db_column='_order')

    objects = DataCategoryManager()

    class Meta(object):
        ordering = ('-parent__id', 'order', 'name')
        verbose_name_plural = 'data categories'


class DataField(BasePlural):
    """Describes the significance and/or meaning behind some data. In addition,
    it defines the natural key of the Django field that represents the location
    of that data e.g. ``library.book.title``.
    """
    app_name = models.CharField(max_length=50)
    model_name = models.CharField(max_length=50)
    field_name = models.CharField(max_length=50)

    internal = models.BooleanField(default=False,
        help_text='Flag for internal use and does not abide by the published' \
            'and archived rules.')

    # An optional unit for this field's data. In some cases databases may have
    # a separate column which denotes the unit for another column, but this is
    # not always the case. Measurement data, for example, should be
    # standardized in the database to allow for consistent querying, thus not
    # requiring a separate column denoting the unit per value.
    unit = models.CharField(max_length=30, null=True, blank=True)
    unit_plural = models.CharField(max_length=40, null=True, blank=True)

    # Although a category does not technically need to be defined, this is more
    # for workflow reasons than for when the concept is published. Automated
    # prcesses may create concepts on the fly, but not know which category they
    # should be linked to initially.
    category = models.ForeignKey(DataCategory, null=True, blank=True)

    # Set this field to true to make this field's values enumerable. This
    # should only be enabled for data that contains a discrete vocabulary, i.e.
    # no full text data.
    enumerable = models.BooleanField(default=False)

    # An optional translator which customizes input query conditions
    # to a format which is suitable for the database.
    translator = models.CharField(max_length=100, blank=True, null=True,
        choices=translators.choices)

    # The timestamp of the last time the underlying data for this field
    # has been modified. This is used for the cache key and for general
    # reference. The underlying table may not have a timestamp nor is it
    # optimal to have to query the data for the max `modified` time.
    data_modified = models.DateTimeField(null=True, help_text='The last time '\
        ' the underlying data for this field was modified.')

    # Certain fields may not be relevant or appropriate for all
    # sites being deployed. This is primarily for preventing exposure of
    # access to private data from certain sites. For example, there may and
    # internal and external deployment of the same site. The internal site
    # has full access to all fields, while the external may have a limited set.
    # NOTE this is not reliable way to prevent exposure of sensitive data.
    # This should be used to simply hide _access_ to the concepts.
    group = models.ForeignKey(Group, null=True, blank=True, related_name='fields+')

    sites = models.ManyToManyField(Site, blank=True, related_name='fields+')

    # The order of this datafield with respect to the category (if defined).
    order = models.FloatField(null=True, blank=True, db_column='_order')

    objects = DataFieldManager()

    class Meta(object):
        unique_together = ('app_name', 'model_name', 'field_name')
        ordering = ('name',)
        permissions = (
            ('view_datafield', 'Can view datafield'),
        )

    def __unicode__(self):
        if self.name:
            return self.name
        if self.lexicon or self.objectset:
            return self.model._meta.verbose_name
        return '{0} {1}'.format(self.model._meta.verbose_name,
            self.field.verbose_name).title()

    def __len__(self):
        return self.size

    def __nonzero__(self):
        "Takes precedence over __len__, so it is always truthy."
        return True

    # The natural key should be used any time fields are being exported
    # for integration in another system. It makes it trivial to map to new
    # data models since there are discrete parts (as suppose to using the
    # primary key).
    def natural_key(self):
        return self.app_name, self.model_name, self.field_name

    # Django Model Field-related Properties and Methods

    @property
    def real_model(self):
        "Returns the model class this datafield is associated with."
        if not hasattr(self, '_real_model'):
            self._real_model = models.get_model(self.app_name, self.model_name)
        return self._real_model

    @property
    def real_field(self):
        "Returns the field object this datafield is associated with."
        if self.real_model:
            try:
                return self.real_model._meta.get_field(self.field_name)
            except FieldDoesNotExist:
                pass

    @property
    def model(self):
        "Returns the model class this datafield represents."
        real_field = self.real_field
        # Handle foreign key fields to a Lexicon model
        if real_field and isinstance(real_field, models.ForeignKey) \
                and issubclass(real_field.rel.to, Lexicon):
            return real_field.rel.to
        return self.real_model

    @property
    def field(self):
        "Returns the field object this datafield represents."
        model = self.model
        if model and issubclass(model, (Lexicon, ObjectSet)):
            return model._meta.pk
        return self.real_field

    @property
    def nullable(self):
        "Returns whether this field can contain NULL values."
        return self.field.null

    @property
    def internal_type(self):
        "Returns the internal type of the field this datafield represents."
        return utils.get_internal_type(self.field)

    @property
    def simple_type(self):
        """Returns a simple type mapped from the internal type."

        By default, it will use the field's internal type, but can be
        overridden by the ``SIMPLE_TYPE_MAP`` setting.
        """
        return SIMPLE_TYPE_MAP.get(self.internal_type, self.internal_type)

    @property
    def lexicon(self):
        """Returns true if the model is a subclass of Lexicon and this
        is the pk field. All other fields on the class are treated as
        normal datafields.
        """
        return self.model and issubclass(self.model, Lexicon) \
            and self.field == self.model._meta.pk

    @property
    def objectset(self):
        """Returns true if the model is a subclass of ObjectSet and this
        is the pk field. All other fields on the class are treated as
        normal datafields.
        """
        return self.model and issubclass(self.model, ObjectSet) \
            and self.field == self.model._meta.pk

    @property
    def searchable(self):
        "Returns true if a text-field and is not an enumerable field."
        return self.simple_type == 'string' and not self.enumerable


    # Convenience Methods
    # Easier access to the underlying data for this data field

    @property
    def values_list(self):
        "Returns a `ValuesListQuerySet` of values for this field."
        if self.lexicon or self.objectset:
            return self.model.objects.values_list('pk', flat=True)
        return self.model.objects.values_list(self.field_name, flat=True)\
            .order_by(self.field_name).distinct()

    def search(self, query):
        "Rudimentary search for string-based values."
        if self.simple_type == 'string' or self.lexicon:
            if self.lexicon:
                field_name = 'value'
            else:
                field_name = self.field_name
            filters = {'{0}__icontains'.format(field_name): query}
            return self.values_list.filter(**filters).iterator()

    def get_plural_unit(self):
        if self.unit_plural:
            plural = self.unit_plural
        elif self.unit and not self.unit.endswith('s'):
            plural = self.unit + 's'
        else:
            plural = self.unit
        return plural

    def get_label(self, value):
        """Gets the label for a particular raw data value.
        If this is classified as `searchable`, a database hit will occur.
        """
        if self.searchable:
            kwargs = {self.field_name: value}
            if self.lexicon:
                return self.model.objects.filter(**kwargs)\
                    .values_list('label', flat=True)[0]
            if self.objectset:
                return self.model.objects.filter(**kwargs)\
                    .values_list('name', flat=True)[0]
            return smart_unicode(value)
        return dict(self.choices)[value]

    # Data-related Cached Properties
    # These may be cached until the underlying data changes

    @cached_property('size', version='data_modified')
    def size(self):
        "Returns the count of distinct values."
        return self.values_list.count()

    @cached_property('values', version='data_modified')
    def values(self):
        "Returns a distinct list of the values."
        return tuple(self.values_list)

    @cached_property('labels', version='data_modified')
    def labels(self):
        """Returns an ordered set of labels corresponding to the values.
        If this field represents to a Lexicon subclass, the `label` field
        will be used, otherwise the values will simply be unicoded.
        """
        if self.lexicon:
            return tuple(self.model.objects.values_list('label', flat=True))
        if self.objectset:
            return tuple(self.model.objects.values_list('name', flat=True))
        # Unicode each value, use an iterator here to prevent loading the
        # raw values in memory
        return map(smart_unicode, iter(self.values_list))

    @cached_property('codes', version='data_modified')
    def codes(self):
        "Returns a distinct set of coded values for this field"
        if self.lexicon:
            return tuple(self.model.objects.values_list('code', flat=True))

    @property
    def choices(self):
        "Returns a distinct set of choices for this field."
        return zip(self.values, self.labels)

    @property
    def coded_choices(self):
        "Returns a distinct set of coded choices for this field."
        if self.lexicon:
            return zip(self.codes, self.labels)

    @property
    def coded_values(self):
        "Returns a distinct set of coded values for this field."
        if self.lexicon:
            return zip(self.values, self.codes)

    # Data Aggregation Properties
    def groupby(self, *args):
        return Aggregator(self.field).groupby(*args)

    def count(self, *args, **kwargs):
        "Returns an the aggregated counts."
        return Aggregator(self.field).count(*args, **kwargs)

    def max(self, *args):
        "Returns the maximum value."
        return Aggregator(self.field).max(*args)

    def min(self, *args):
        "Returns the minimum value."
        return Aggregator(self.field).min(*args)

    def avg(self, *args):
        "Returns the average value. Only applies to quantitative data."
        if self.simple_type == 'number':
            return Aggregator(self.field).avg(*args)

    def sum(self, *args):
        "Returns the sum of values. Only applies to quantitative data."
        if self.simple_type == 'number':
            return Aggregator(self.field).sum(*args)

    def stddev(self, *args):
        "Returns the standard deviation. Only applies to quantitative data."
        if self.simple_type == 'number':
            return Aggregator(self.field).stddev(*args)

    def variance(self, *args):
        "Returns the variance. Only applies to quantitative data."
        if self.simple_type == 'number':
            return Aggregator(self.field).variance(*args)

    # Translator Convenience Methods

    @property
    def operators(self):
        "Returns the valid operators for this datafield."
        trans = translators[self.translator]
        return [(x, operators[x].verbose_name) for x in trans.get_operators(self)]

    def translate(self, operator=None, value=None, tree=None, **context):
        "Convenince method for performing a translation on a query condition."
        trans = translators[self.translator]
        return trans.translate(self, operator, value, tree, **context)

    def validate(self, operator=None, value=None, tree=None, **context):
        "Convenince method for performing a translation on a query condition."
        trans = translators[self.translator]
        return trans.validate(self, operator, value, tree, **context)


class DataConcept(BasePlural):
    """Our acceptance of an ontology is, I think, similar in principle to our
    acceptance of a scientific theory, say a system of physics; we adopt, at
    least insofar as we are reasonable, the simplest conceptual scheme into
    which the disordered fragments of raw experience can be fitted and
    arranged.

        -- Willard Van Orman Quine
    """

    ident = models.CharField(max_length=100, null=True,
        blank=True, validators=[validate_ident],
        help_text='Unique identifier that can be used for programmatic access')

    internal = models.BooleanField(default=False,
        help_text='Flag for internal use and does not abide by the published' \
            'and archived rules.')

    # Although a category does not technically need to be defined, this more
    # for workflow reasons than for when the concept is published. Automated
    # prcesses may create concepts on the fly, but not know which category they
    # should be linked to initially. the admin interface enforces choosing a
    # category when the concept is published
    category = models.ForeignKey(DataCategory, null=True, blank=True)

    # The associated fields for this concept. fields can be
    # associated with multiple concepts, thus the M2M
    fields = models.ManyToManyField(DataField, through='DataConceptField',
        related_name='concepts')

    # Certain concepts may not be relevant or appropriate for all
    # sites being deployed. This is primarily for preventing exposure of
    # access to private data from certain sites. For example, there may and
    # internal and external deployment of the same site. The internal site
    # has full access to all fields, while the external may have a limited set.
    # NOTE this is not reliable way to prevent exposure of sensitive data.
    # This should be used to simply hide _access_ to the concepts.
    group = models.ForeignKey(Group, null=True, blank=True,
        related_name='concepts+')

    sites = models.ManyToManyField(Site, blank=True, related_name='concepts+')

    order = models.FloatField(null=True, blank=True, db_column='_order')

    # An optional formatter which provides custom formatting for this
    # concept relative to the associated fields. If a formatter is not
    # defined, this DataConcept is not intended to be exposed since the
    # underlying data may not be appropriate for client consumption.
    formatter_name = models.CharField('formatter', max_length=100, blank=True,
        null=True, choices=formatters.choices)

    queryview = models.CharField(max_length=100, blank=True, null=True,
        choices=queryviews.choices)

    # A flag that denotes when this concept can be applied to an ORDER BY
    # Certain concepts are not appropriate because they are too complicated,
    # or a very specific abstraction that does not order by what it actually
    # represents.
    sortable = models.BooleanField(default=True)

    objects = DataConceptManager()

    def format(self, *args, **kwargs):
        """Convenience method for formatting data relative to this concept's
        associated formatter. To prevent redundant initializations (say, in
        a tight loop) the formatter instance is cached until the formatter
        name changes.
        """
        name = self.formatter_name
        cache = getattr(self, '_formatter_cache', None)
        if not cache or name != cache[0]:
            formatter = formatters.get(name)(self)
            self._formatter_cache = (name, formatter)
        else:
            formatter = cache[1]
        return formatter(*args, **kwargs)

    class Meta(object):
        app_label = 'avocado'
        ordering = ('order',)
        permissions = (
            ('view_dataconcept', 'Can view dataconcept'),
        )


class DataConceptField(models.Model):
    "Through model between DataConcept and DataField relationships."
    field = models.ForeignKey(DataField, related_name='concept_fields')
    concept = models.ForeignKey(DataConcept, related_name='concept_fields')

    name = models.CharField(max_length=100, null=True, blank=True)
    name_plural = models.CharField(max_length=100, null=True, blank=True)
    order = models.FloatField(null=True, blank=True, db_column='_order')

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta(object):
        ordering = ('order',)

    def __unicode__(self):
        return unicode(self.name or self.field.name)

    def get_plural_name(self):
        return self.name_plural or self.field.get_plural_name()


class DataContext(AbstractDataContext, Base):
    """JSON object representing one or more data field conditions. The data may
    be a single condition, an array of conditions or a tree stucture.

    This corresponds to the `WHERE` statements in a SQL query.
    """
    session = models.BooleanField(default=False)

    # For authenticated users the `user` can be directly referenced,
    # otherwise the session key can be used.
    user = models.ForeignKey(User, null=True, blank=True, related_name='datacontext+')
    session_key = models.CharField(max_length=40, null=True, blank=True)

    def __unicode__(self):
        if self.user_id:
            out = unicode(self.user)
        else:
            out = self.session_key
        return u'{0} ({1})'.format(out, 'session' if self.session else 'archived')

    def archive(self):
        if self.archived:
            return False
        backup = self.pk, self.session, self.archived
        self.pk, self.session, self.archived = None, False, True
        self.save()
        self.pk, self.session, self.archived = backup
        return True


class DataView(AbstractDataView, Base):
    """JSON object representing one or more data field conditions. The data may
    be a single condition, an array of conditions or a tree stucture.

    This corresponds to the `SELECT` and `ORDER BY` statements in a SQL query.
    """
    session = models.BooleanField(default=False)

    # For authenticated users the `user` can be directly referenced,
    # otherwise the session key can be used.
    user = models.ForeignKey(User, null=True, blank=True, related_name='dataview+')
    session_key = models.CharField(max_length=40, null=True, blank=True)

    def __unicode__(self):
        if self.user_id:
            out = unicode(self.user)
        else:
            out = self.session_key
        return u'{0} ({1})'.format(out, 'session' if self.session else 'archived')

    def archive(self):
        if self.archived:
            return False
        backup = self.pk, self.session, self.archived
        self.pk, self.session, self.archived = None, False, True
        self.save()
        self.pk, self.session, self.archived = backup
        return True


# Register instance-level cache invalidation handlers
post_save.connect(post_save_cache, sender=DataField)
post_save.connect(post_save_cache, sender=DataConcept)
post_save.connect(post_save_cache, sender=DataCategory)

pre_delete.connect(pre_delete_uncache, sender=DataField)
pre_delete.connect(pre_delete_uncache, sender=DataConcept)
pre_delete.connect(pre_delete_uncache, sender=DataCategory)



import re
import functools
from django.dispatch import receiver
from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import LazyObject
from django.conf import settings as django_settings
from django.test.signals import setting_changed
from avocado.conf import global_settings


class Settings(object):
    def __init__(self, settings_dict):
        # set the initial settings as defined in the global_settings
        for setting in iter(dir(global_settings)):
            setattr(self, setting, getattr(global_settings, setting))

        # iterate over the user-defined settings and override the default
        # settings
        for key, value in iter(settings_dict.items()):
            setattr(self, key, value)

    def __setattr__(self, key, value):
        if key == key.upper():
            default = getattr(self, key, None)
            if isinstance(default, dict):
                default.update(value)
            else:
                object.__setattr__(self, key, value)


class LazySettings(LazyObject):
    def _setup(self):
        self._wrapped = Settings(getattr(django_settings, 'AVOCADO', {}))


settings = LazySettings()

@receiver(setting_changed)
def test_setting_changed_handler(**kwargs):
    if kwargs['setting'] == 'AVOCADO':
        for key, value in kwargs['value'].iteritems():
            setattr(settings._wrapped, key, value)
    elif kwargs['setting'].startswith('AVOCADO_'):
        key = kwargs['setting'][8:]
        value = kwargs['value']
        setattr(settings._wrapped, key, value)


class Dependency(object):
    name = ''

    def __nonzero__(self):
        return self.installed and self.setup

    def __unicode__(self):
        return self.name

    @property
    def doc(self):
        return re.sub('\n( )+', '\n', self.__doc__ or '').strip('\n')

    @property
    def marked_doc(self):
        try:
            import markdown
        except ImportError:
            return self.doc
        return markdown.markdown(self.doc)

    def test_install(self):
        raise NotImplemented

    def test_setup(self):
        return self.test_install()

    @property
    def installed(self):
        return self.test_install() is not False

    @property
    def setup(self):
        return self.test_setup() is not False


class Haystack(Dependency):
    """What's having all this great descriptive data if no one can find it?
    Haystack provides search engine facilities for the metadata.

    Install by doing `pip install django-haystack` and installing one of the
    supported search engine backends. The easiest to setup is Whoosh which is
    implemented in pure Python. Install it by doing `pip install whoosh`. Add
    haystack to `INSTALLED_APPS`.
    """

    name = 'haystack'

    # An import may cause an improperly configured error, so this catches
    # and returns the error for downstream use.
    def _test(self):
        try:
            import haystack
        except (ImportError, ImproperlyConfigured), e:
            return e

    def test_install(self):
        error = self._test()
        if isinstance(error, ImportError):
            return False

    def test_setup(self):
        error = self._test()
        if isinstance(error, ImproperlyConfigured):
            return False


#class Numpy(Dependency):
#    name = 'numpy'
#
#    def test_install(self):
#        try:
#            import numpy
#        except ImportError:
#            return False


class Scipy(Dependency):
    """Avocado comes with a stats package for performing some rudimentary
    statistical, aggregation and clustering operations on the data. This is not
    always required or necessary for all data, but if there is a heavy emphasis
    on numerical data or the amount of data is quite large, the stats may come
    in handy.

    Install by doing `pip install numpy` first (a dependency of SciPy), followed
    by `pip install scipy`. Note, there are a few dependencies for compilation,
    so review SciPy's [installation instructions](http://www.scipy.org/Installing_SciPy)
    for more details.
    """

    name = 'scipy'

    def test_install(self):
        try:
            import scipy
        except ImportError:
            return False


class Openpyxl(Dependency):
    """Avocado comes with an export package for supporting various means of
    exporting data into different formats. One of those formats is the native
    Microsoft Excel .xlsx format. To support that, the openpyxl library is used.

    Install by doing `pip install openpyxl`.
    """

    name = 'openpyxl'

    def test_install(self):
        try:
            import openpyxl
        except ImportError:
            return False


class Guardian(Dependency):
    """This enables fine-grain control over who has permissions for various
    DataFields. Permissions can be defined at a user or group level.

    Install by doing `pip install django-guardian` and adding guardian to
    `INSTALLED_APPS`.
    """

    name = 'guardian'

    def test_install(self):
        try:
            import guardian
        except ImportError:
            return False


# Keep track of the officially supported apps and libraries used for various
# features.
OPTIONAL_DEPS = {
    'haystack': Haystack(),
#    'numpy': Numpy(),
    'scipy': Scipy(),
    'openpyxl': Openpyxl(),
    'guardian': Guardian(),
}


def requires_dep(lib):
    "Decorator for functions that require a supported third-party library."
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            dep = OPTIONAL_DEPS[lib]
            if not dep:
                raise ImproperlyConfigured('{0} must be installed to use this feature.\n\n{1}'.format(lib, dep.__doc__))
            return f(*args, **kwargs)
        return wrapper
    return decorator



from optparse import make_option
from django.core.management.base import BaseCommand
from django.db import transaction
from django.template import loader, Context
from django.conf import settings as django_settings
from avocado.models import DataField
from avocado.conf import OPTIONAL_DEPS
from modeltree.tree import trees


class Command(BaseCommand):
    help = """Performs a series of checks for the setup and installation as well
    as checks for any orphaned data fields.
    """

    __doc__ = help

    option_list = BaseCommand.option_list + (
        make_option('--output', default='stdout',
            help='Specify the output type: stdout, html'),
    )

    @transaction.commit_on_success
    def handle(self, *args, **options):
        output = options.get('output')

        unknown_models = []
        unknown_fields = []


        for f in DataField.objects.iterator():
            if f.model is None:
                unknown_models.append(f)
            elif f.field is None:
                unknown_fields.append(f)

        # Ensure the default tree is valid
        try:
            trees.default
            default_modeltree = trees.modeltrees['default']['model']
        except Exception:
            default_modeltree = None

        context = {
            'unknown_models': unknown_models,
            'unknown_fields': unknown_fields,
            'optional_deps': OPTIONAL_DEPS,
            'settings': getattr(django_settings, 'AVOCADO', {}),
            'export_package_installed': 'avocado.export' in django_settings.INSTALLED_APPS,
            'default_modeltree': default_modeltree,
        }

        if output == 'stdout':
            template = loader.get_template('avocado/check.txt')
        elif output == 'html':
            template = loader.get_template('avocado/check.html')

        return template.render(Context(context))



import os
import sys
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from optparse import make_option
from django.db.models import (get_model, get_models, get_app, AutoField,
    ForeignKey, OneToOneField, ManyToManyField, FieldDoesNotExist)
from django.core.management.base import BaseCommand
from avocado.models import DataField
from avocado.lexicon.models import Lexicon
from avocado.sets.models import ObjectSet
from avocado.core import utils


class Command(BaseCommand):
    """
    SYNOPSIS::

        python manage.py avocado init [options...] labels

    DESCRIPTION:

        Finds all models referenced by the app, model or field `labels` and
        attempts to create a `DataField` instance per model field.
        Any `DataField` already loaded will not be altered in any way.

    OPTIONS:

        `--include-non-editable` - Create `DataField` instances for fields
        marked as not editable (i.e. `editable=False`).

        `--include-keys` - Create `DataField` instances for primary key
        and foreign key fields.

        `--update` - Updates existing `DataField` instances with metadata from
        model fields. Note this overwrites any descriptive metadata changes made
        to `DataField` such as `name`, `name_plural`, and `description`.

        `--force` - Synonym to `--update`
    """

    help = '\n'.join([
        'Finds all models referenced by the app, model or field `labels` and',
        'attempts to create a `DataField` instance per model field.',
        'Any `DataField` already loaded will not be altered in any way.'
    ])

    args = 'app [app.model, [app.model.field, [...]]]'

    option_list = BaseCommand.option_list + (
        make_option('-e', '--include-non-editable', action='store_true',
            dest='include_non_editable', default=False,
            help='Create fields for non-editable fields'),

        make_option('-k', '--include-keys', action='store_true',
            dest='include_keys', default=False,
            help='Create fields for primary and foreign key fields'),

        make_option('-f', '--force', action='store_true',
            dest='force', default=False,
            help='Forces an update on existing field metadata'),

        make_option('-q', '--quiet', action='store_true',
            dest='quiet', default=False,
            help='Do not print any output')
    )

    # These are ignored since these join fields will be determined at runtime
    # using the modeltree library. fields can be created for any other
    # these field types manually
    key_field_types = (
        AutoField,
        ForeignKey,
        OneToOneField,
    )

    ignored_lexicon_fields = ('value', 'order', 'label', 'code')

    def handle(self, *args, **options):
        "Handles app_label or app_label.model_label formats."

        if options.get('quiet'):
            self.stdout = sys.stdout
            sys.stdout = open(os.devnull, 'w')

        if options.get('force'):
            resp = raw_input('Forcing a init will update metadata for '
                'existing fields. Are you sure you want to do this?\n'
                'This will overwrite any previous changes made. Type "yes" '
                'to continue: ')
            if resp.lower() != 'yes':
                print 'Initialization operation cancelled'
                return

        for label in args:
            pending_fields = []
            pending_models = []

            toks = label.split('.')
            app_name = model_name = field_name = None

            # Specific field
            if len(toks) == 3:
                app_name, model_name, field_name = toks
            # All fields for a model
            elif len(toks) == 2:
                app_name, model_name = toks
            # All fields for each model in the app
            else:
                app_name = toks[0]

            if model_name:
                model = get_model(app_name, model_name)

                if model is None:
                    print 'Cannot find model "{0}", skipping...'.format(label)
                    continue

                # Specific field
                if field_name:
                    try:
                        field = model._meta.get_field_by_name(field_name)[0]
                    except FieldDoesNotExist:
                        print 'Cannot find field "{0}", skipping...'.format(label)
                        continue
                    pending_fields = [(field, model_name, app_name)]

                # No specific field, queue up the model
                else:
                    pending_models.append(model)
            else:
                app = get_app(app_name)
                if app is None:
                    print 'Cannot find app "{0}", skipping...'.format(label)
                    continue
                pending_models.extend(get_models(app))

            for model in pending_models:
                lexicon = issubclass(model, Lexicon)
                objectset = issubclass(model, ObjectSet)

                model_name = model._meta.object_name.lower()

                if lexicon or objectset:
                    pk = model._meta.pk
                    pk.verbose_name = model._meta.verbose_name
                    pending_fields.append((pk, model_name, app_name))
                else:
                    for field in model._meta.fields:
                        pending_fields.append((field, model_name, app_name))

            added = 0
            updated = 0

            for field_args in pending_fields:
                status = self.handle_field(*field_args, **options)
                if status is True:
                    added += 1
                elif status is False:
                    updated += 1

            if added == 1:
                print '1 field added for {0}'.format(label)
            elif added > 1:
                print '{0} fields added for {1}'.format(added, label)

            if updated == 1:
                print '1 field updated for {0}'.format(label)
            elif updated > 1:
                print '{0} fields updated for {1}'.format(updated, label)

        if options.get('quiet'):
            sys.stdout = self.stdout

    def handle_field(self, field, model_name, app_name, **options):
        include_keys = options.get('include_keys')
        force = options.get('force')
        include_non_editable = options.get('include_non_editable')

        # M2Ms do not make any sense here..
        if isinstance(field, ManyToManyField):
            return

        # Lexicons and ObjectSets are represented via their primary key, so
        # these may pass
        if not issubclass(field.model, (Lexicon, ObjectSet)):
            # Check for primary key, and foreign key fields
            if isinstance(field, self.key_field_types) and not include_keys:
                return

            # Ignore non-editable fields since in most cases they are for
            # managment purposes
            if not field.editable and not include_non_editable:
                return

        # All but the field name is case-insensitive, do initial lookup
        # to see if it already exists, skip if it does
        lookup = {
            'app_name__iexact': app_name,
            'model_name__iexact': model_name,
            'field_name': field.name,
        }

        # Note, `name` is set below
        kwargs = {
            'description': field.help_text or None,
            'app_name': app_name,
            'model_name': model_name.lower(),
            'field_name': field.name,
        }

        try:
            datafield = DataField.objects.get(**lookup)
        except DataField.DoesNotExist:
            datafield = DataField(published=False, **kwargs)

        if datafield.pk:
            created = False
            if not force:
                print '({0}) {1}.{2} already exists. Skipping...'.format(app_name,
                    model_name, field.name)
                return
            # Only overwrite if the source value is not falsy
            datafield.__dict__.update([(k, v) for k, v in kwargs.items()])
        else:
            created = True

        if not datafield.name:
            # Use the default unicode representation of the datafield
            datafield.name = unicode(datafield)

        # Update fields with flags
        datafield.__dict__.update(utils.get_heuristic_flags(datafield))
        datafield.save()
        return created



from django.db import models
from django.contrib.auth.models import Group
from django.contrib.sites.models import Site


# Field...

class Field(models.Model):
    app_name = models.CharField(max_length=100)
    model_name = models.CharField(max_length=100)
    field_name = models.CharField(max_length=100)

    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)
    keywords = models.CharField(max_length=100, null=True, blank=True)

    is_public = models.BooleanField(default=False)
    group = models.ForeignKey(Group, null=True, blank=True)
    sites = models.ManyToManyField(Site, blank=True)

    translator = models.CharField(max_length=100, blank=True, null=True)

    enable_choices = models.BooleanField(default=False)

    choices_handler = models.TextField(null=True, blank=True, help_text="""
        Allowed callbacks include specifying:
            1. a constant name on the model
            2. a constant name on the model's module
            3. a string that can be evaluated
    """)

    class Meta(object):
        db_table = 'avocado_field'

    def __unicode__(self):
        return unicode(self.name)

    def natural_key(self):
        return (self.app_name, self.model_name, self.field_name)


# Category...

class Category(models.Model):
    name = models.CharField(max_length=100)
    parent = models.ForeignKey('self', null=True, blank=True)
    order = models.FloatField(default=0)

    class Meta(object):
        db_table = 'avocado_category'

    def __unicode__(self):
        return unicode(self.name)


# Concepts...

class Concept(models.Model):
    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)
    keywords = models.CharField(max_length=100, null=True, blank=True)
    category = models.ForeignKey(Category, null=True, blank=True)
    is_public = models.BooleanField(default=False)
    order = models.FloatField(default=0)

    class Meta(object):
        abstract = True

    def __unicode__(self):
        return unicode(self.name)


class ConceptField(models.Model):
    name = models.CharField(max_length=100, null=True, blank=True)
    order = models.FloatField(default=0)

    class Meta(object):
        abstract = True
        ordering = ('order',)

    def __unicode__(self):
        return unicode(self.field)


class Criterion(Concept):
    fields = models.ManyToManyField(Field, through='CriterionField')
    viewset = models.CharField(max_length=100)

    class Meta(Concept.Meta):
        db_table = 'avocado_criterion'


class CriterionField(ConceptField):
    concept = models.ForeignKey(Criterion, related_name='conceptfields')
    field = models.ForeignKey(Field, limit_choices_to={'is_public': True})
    required = models.BooleanField(default=True)

    class Meta(ConceptField.Meta):
        db_table = 'avocado_criterionfield'


class Column(Concept):
    "An interface to specify the necessary fields for a column."
    fields = models.ManyToManyField(Field, through='ColumnField')
    html_fmtr = models.CharField(max_length=100)
    csv_fmtr = models.CharField(max_length=100)

    class Meta(Concept.Meta):
        db_table = 'avocado_column'


class ColumnField(ConceptField):
    concept = models.ForeignKey(Column, related_name='conceptfields')
    field = models.ForeignKey(Field)

    class Meta(ConceptField.Meta):
        db_table = 'avocado_columnfield'



from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from avocado.models import DataField


class DataFieldAdminForm(forms.ModelForm):
    def clean_app_name(self):
        app_name = self.cleaned_data.get('app_name')
        try:
            models.get_app(app_name)
        except ImproperlyConfigured:
            raise forms.ValidationError('The app "{0}" could not be found'.format(app_name))
        return app_name

    def clean(self):
        cleaned_data = self.cleaned_data
        app_name = self.cleaned_data.get('app_name')
        model_name = cleaned_data.get('model_name')
        field_name = cleaned_data.get('field_name')

        model = models.get_model(app_name, model_name)
        if model is None:
            del cleaned_data['model_name']
            msg = 'The model "{0}" could not be found in the app "{1}"'.format(model_name, app_name)
            self._errors['model_name'] = self.error_class([msg])
        elif not model._meta.get_field_by_name(field_name):
            del cleaned_data['field_name']
            msg = 'The model "{0}" does not have a field named "{1}"'.format(model_name, field_name)
            self._errors['field_name'] = self.error_class([msg])
        return cleaned_data

    class Meta(object):
        model = DataField



from haystack import site
from haystack.indexes import *
from avocado.models import DataConcept, DataField


class DataConceptIndex(SearchIndex):
    text = CharField(document=True, use_template=True)

    def get_model(self):
        return DataConcept


class DataFieldIndex(SearchIndex):
    text = CharField(document=True, use_template=True)

    def get_model(self):
        return DataField


site.register(DataField, DataFieldIndex)
site.register(DataConcept, DataConceptIndex)



from django.http import HttpResponse
from django.core.exceptions import ImproperlyConfigured
from avocado.conf import OPTIONAL_DEPS
if not OPTIONAL_DEPS['openpyxl']:
    raise ImproperlyConfigured('openpyxl must be installed to use this exporter.')

from openpyxl import Workbook
from _base import BaseExporter


class ExcelExporter(BaseExporter):
    short_name = 'Excel'
    long_name = 'Microsoft Excel 2007 Format'

    file_extension = 'xlsx'
    content_type = 'application/vnd.ms-excel'

    preferred_formats = ('excel', 'boolean', 'number', 'string')

    def write(self, iterable, buff=None, *args, **kwargs):
        buff = self.get_file_obj(buff)

        wb = Workbook(optimized_write=True)

        ws_data = wb.create_sheet()
        ws_data.title = 'Data'

        header = []
        # Create the data worksheet
        for i, row_gen in enumerate(self.read(iterable, *args, **kwargs)):
            row = []
            for data in row_gen:
                if i == 0:
                    # Build up header row
                    header.extend(data.keys())
                # Add formatted section to the row
                row.extend(data.values())
            # Write headers on first iteration
            if i == 0:
                ws_data.append(header)
            ws_data.append(row)

        ws_dict = wb.create_sheet()
        ws_dict.title = 'Data Dictionary'

        # Create the Data Dictionary Worksheet
        ws_dict.append(('Field Name', 'Data Type', 'Description',
            'Concept Name', 'Concept Discription'))

        for c in self.concepts:
            cfields = c.concept_fields.select_related('field')
            for cfield in cfields:
                field = cfield.field
                ws_dict.append((field.field_name, field.simple_type,
                    field.description, c.name, c.description))

        # This hacked up implementation is due to `save_virtual_workbook`
        # not behaving correctly. This function should handle the work
        # https://bitbucket.org/ericgazoni/openpyxl/src/94b05cf9defb9787b4dfbf9e8dca7ba6e0b33d56/openpyxl/writer/excel.py?at=default#cl-154
        # however, no data is actually being saved to the worksheets..
        if isinstance(buff, HttpResponse):
            _buff = self.get_file_obj()
            wb.save(_buff)
            buff.content = _buff.getvalue()
            _buff.close()
        else:
            wb.save(buff)
        return buff
